#include "map.hpp"
#include "dbc.hpp"
#include <vector>
#include <fmt/core.h>
#include <random>
#include <utility>


std::random_device g_rng;
std::mt19937 g_generator(g_rng());

using std::vector, std::pair;
using namespace fmt;

void dump_map(const std::string &msg, Matrix &map) {
  println("----------------- {}", msg);
  for(auto row : map) {
    for(auto col : row) {
      print("{} ", col);
    }
    print("\n");
  }
}

inline void add_neighbors(PairList &neighbors, Matrix &closed, size_t j, size_t i) {
  size_t h = closed.size();
  size_t w = closed[0].size();
  vector<size_t> rows{j - 1, j, j + 1};
  vector<size_t> cols{i - 1, i, i + 1};

  for(auto row : rows) {
    for(auto col : cols) {
      if((0 <= row && row < h) &&
          (0 <= col && col < w) &&
          closed[row][col] == 0)
      {
        closed[row][col] = 1;
        neighbors.push_back({.j=row, .i=col});
      }
    }
  }
}

Map::Map(size_t width, size_t height) : m_limit(1000) {
  m_walls = Matrix(height, MatrixRow(width, 0));
  m_input_map = Matrix(height, MatrixRow(width, 1));
}

void Map::make_paths() {
  size_t h = m_input_map.size();
  size_t w = m_input_map[0].size();

  // Initialize the new array with every pixel at limit distance
  // NOTE: this is normally ones() * limit
  int limit = m_limit == 0 ? h * w : m_limit;
  Matrix new_arr = Matrix(h, MatrixRow(w, limit));
  Matrix closed = m_walls;
  PairList starting_pixels;
  PairList open_pixels;

  // First pass: Add starting pixels and put them in closed
 for(size_t counter = 0; counter < h * w; counter++) {
    size_t i = counter % w;
    size_t j = counter / w;
    if(m_input_map[j][i] == 0) {
      new_arr[j][i] = 0;
      closed[j][i] = 1;
      starting_pixels.push_back({.j=j,.i=i});
    }
  }

  // Second pass: Add border to open
  for(auto sp : starting_pixels) {
    add_neighbors(open_pixels, closed, sp.j, sp.i);
  }

  // Third pass: Iterate filling in the open list
  int counter = 1; // leave this here so it's available below
  for(; counter < limit && !open_pixels.empty(); ++counter) {
    PairList next_open;
    for(auto sp : open_pixels) {
      new_arr[sp.j][sp.i] = counter;
      add_neighbors(next_open, closed, sp.j, sp.i);
    }
    open_pixels = next_open;
  }

  // Last pass: flood last pixels
  for(auto sp : open_pixels) {
    new_arr[sp.j][sp.i] = counter;
  }

  m_paths = new_arr;
}

void Map::make_room(size_t origin_x, size_t origin_y, size_t w, size_t h) {
  dbc::pre("x out of bounds", origin_x < width());
  dbc::pre("y out of bounds", origin_y < height());
  dbc::pre("w out of bounds", w <= width());
  dbc::pre("h out of bounds", h <= height());

  for(size_t y = origin_y; y < origin_y + h; ++y) {
    dbc::check(y < m_walls.size(), "y is out of bounds");
    for(size_t x = origin_x; x < origin_x + w; ++x) {
      dbc::check(x < m_walls[y].size(), "x is out of bounds");
      m_walls[y][x] = 1;
    }
  }
}

inline int make_split(Room &cur, bool horiz) {
  println("MAKE SPLIT horiz={}, y={}, w={}, h={}", horiz,
        cur.y, cur.width, cur.height);
  size_t dimension = horiz ? cur.height : cur.width;
  int min = dimension / 4;
  int max = dimension - min;
  println("dimension={}, min={}, max={}", dimension, min, max);
  std::uniform_int_distribution<int> rand_dim(min, max);
  return rand_dim(g_generator);
}

void Map::partition_map(Room &cur, int depth) {
  if(cur.width >= 5 && cur.width <= 10 &&
      cur.height >= 5 && cur.height <= 10) {
    m_rooms.push_back(cur);
    return;
  }

  std::uniform_int_distribution<int> rsplit(0, 1);
  bool horiz = cur.width > cur.height ? false : true;
  int split = make_split(cur, horiz);
  Room left = cur;
  Room right = cur;

  if(horiz) {
    dbc::check(split > 0, "split is not > 0");
    dbc::check(split < int(cur.height), "split is too big!");
    left.height = size_t(split - 1);
    right.y = cur.y + split;
    right.height = size_t(cur.height - split);
  } else {
    dbc::check(split > 0, "split is not > 0");
    dbc::check(split < int(cur.width), "split is too big!");

    left.width = size_t(split-1);
    right.x = cur.x + split,
    right.width = size_t(cur.width - split);
  }

  if(depth > 0  && left.width > 5 && left.height > 5) {
    partition_map(left, depth-1);
  }

  if(depth > 0 && right.width > 5 && right.height > 5) {
    partition_map(right, depth-1);
  }
}

void Map::place_rooms(Room &cur) {
  for(auto &cur : m_rooms) {
    make_room(cur.x+1, cur.y+1, cur.width-2, cur.height-2);
    cur.door_x = cur.x+1;
    cur.door_y = cur.y;
    m_input_map[cur.door_y][cur.door_x] = 0;
  }
}

void Map::generate() {
  Room root{
    .x = 0,
    .y = 0,
    .width = width(),
    .height = height()
  };

  partition_map(root, 10);
  place_rooms(root);
  make_paths();
  Room &room0 = m_rooms[0];
  Room &room1 = m_rooms[1];

  int cur = m_paths[room0.door_y][room0.door_x];
  int next = m_paths[room0.door_y][room0.door_x+1];
  int i = 1;

  while(next >= cur) {
    cur = next;
    next = m_paths[room0.door_y][room0.door_x+i];
    ++i;
    println("door_y: {}, door_x: {}, CUR: {}, NEXT: {}", room0.door_y, room0.door_x, cur, next);
  }
}


void Map::dump() {
  dump_map("PATHS", m_paths);
  dump_map("WALLS", m_walls);
  dump_map("INPUT", m_input_map);
}
